import numpy as np
import example_robot_data as erd
import pinocchio as pin


class problemDataAbstract:
    """Parameters of an abstract optimal control problem

    Attributes:
        dt (float): time step of the OCP
        dt_wbc (float): time step of the low level integration
        mpc_wbc_ratio (int): ratio between OCP and integration time steps
        T (int): number of running nodes in the OCP
        robot (RobotWrapper): wrapper containing data about the robot
        q0 (vector): default configuration of the robot
        model (Model): model of the robot
        data (Data): data related to the model of the robot
        collision_model (GeometryModel): collision mesh of the robot
        visual_model (GeometryModel): visual mesh of the robot
        frozen_names (list): names of frozen joints
        frozen_idxs (list): IDs of frozen joints
        nq (int): size of the configuration vector
        nv (int): size of the velocity vector
        nx (int): size of the state vector
        ndx (int): size of the derivative of the state vector
        nu (int): size of the control vector
        ntau (int): size of the effort vector
        v0 (vector): initial velocity vector
        x0 (vector): initial state vector
        u0 (vector): initial control vector
        baumgarte_gains (array): Baumgarte gains for the contacts
        xref (vector): reference state
        uref (vector): reference control
    """

    def __init__(self, param, frozen_names=[]):
        self.dt = param.dt_mpc  # OCP dt
        self.dt_wbc = param.dt_wbc
        self.mpc_wbc_ratio = int(self.dt / self.dt_wbc)
        # self.init_steps = 10
        # self.target_steps = param.gait.shape[0]
        self.T = param.gait.shape[0] - 1

        self.robot = erd.load("solo12")
        self.q0 = self.robot.q0
        self.q0[:7] = np.array([0.0, 0.0, 0.2607495, 0, 0, 0, 1])
        self.q0[7:] = param.q_init

        self.model = self.robot.model
        self.rdata = self.model.createData()
        self.collision_model = self.robot.collision_model
        self.visual_model = self.robot.visual_model

        self.frozen_names = frozen_names
        if frozen_names != []:
            self.frozen_idxs = [self.model.getJointId(id) for id in frozen_names]
            self.freeze()

        self.nq = self.model.nq
        self.nv = self.model.nv
        self.nx = self.nq + self.nv
        self.ndx = 2 * self.nv
        self.nu = (
            12 - len(frozen_names) + 1 if len(frozen_names) != 0 else 12
        )  # -1 to take into account the freeflyer
        self.ntau = self.nv

        self.effort_limit = np.ones(self.nu) * 2.5

        self.v0 = np.zeros(18)
        self.x0 = np.concatenate([self.q0, self.v0])
        self.u0 = np.zeros(self.nu)

        self.baumgarte_gains = np.array([0, 100])

        self.xref = self.x0
        self.uref = self.u0

        self.lfFoot, self.rfFoot, self.lhFoot, self.rhFoot = (
            "FL_FOOT",
            "FR_FOOT",
            "HL_FOOT",
            "HR_FOOT",
        )
        self.cnames = [self.lfFoot, self.rfFoot, self.lhFoot, self.rhFoot]
        self.allContactIds = [self.model.getFrameId(f) for f in self.cnames]
        self.lfFootId = self.model.getFrameId(self.lfFoot)
        self.rfFootId = self.model.getFrameId(self.rfFoot)
        self.lhFootId = self.model.getFrameId(self.lhFoot)
        self.rhFootId = self.model.getFrameId(self.rhFoot)
        self.Rsurf = np.eye(3)

    def freeze(self):
        geom_models = [self.visual_model, self.collision_model]
        self.model, geometric_models_reduced = pin.buildReducedModel(
            self.model,
            list_of_geom_models=geom_models,
            list_of_joints_to_lock=self.frozen_idxs,
            reference_configuration=self.q0,
        )
        self.rdata = self.model.createData()
        self.visual_model = geometric_models_reduced[0]
        self.collision_model = geometric_models_reduced[1]


class ProblemData(problemDataAbstract):
    """Parameters of the optimal control problem
    Base is not fixed, with a base velocity tracking task

    Attributes:
        useFixedBase (bool): whether the base is fixed
        baseId (int): ID of the base frame
        mu (float): friction coefficient
        base_velocity_tracking_w (float): weight of the base velocity tracking task
        ground_collision_w (float): weight of the ground collision task
        friction_cone_w (float): weight of the friction cone constraint
        impact_altitude_w (float): weight of the impact altitude constraint
        impact_velocity_w (float): weight of the impact velocity constraint
        control_bound_w (float): weight on the control bound contraint
        control_reg_w (float): weight of the control regularization
        state_reg_w (float): weight of the state regularization
        terminal_velocity_w (float): weight on the terminal velocity value
        force_reg_w (float): weight of the contact force regularization
        xref (vector): reference state
        uref (vector): reference control
    """

    def __init__(self, param):
        super().__init__(param)

        self.useFixedBase = 0
        self.baseId = self.model.getFrameId("base_link")
        self.mu = 0.7

        # Cost function weights
        self.base_velocity_tracking_w = 1e2
        self.ground_collision_w = 1e3 * 0

        # Constraints weights
        self.friction_cone_w = 1e3
        self.impact_altitude_w = 1e3 * 0
        self.impact_velocity_w = 1e3 * 0
        self.control_bound_w = 1e3

        # Regularization weights
        self.control_reg_w = 1e0 * 0
        self.state_reg_w = np.array(
            [0] * 3
            + [1e1] * 3
            + [1e1] * 12
            + [0] * 6
            + [1e1] * 3
            + [1e1] * 3
            + [1e1] * 6
        ) * 0
        self.terminal_velocity_w = np.array([0] * self.nv + [1e4] * self.nv) * 0
        self.force_reg_w = 1e0 * 0

        # Reference values
        self.xref = self.x0
        self.uref = self.u0


class ProblemDataFull(problemDataAbstract):
    """Parameters of the optimal control problem
    Base is fixed, with a foot tracking task

    Attributes:
        useFixedBase (bool): whether the base is fixed
        mu (float): friction coefficient
        foot_tracking_w (float): weight of the foot tracking task
        friction_cone_w (float): weight of the friction cone constraint
        control_bound_w (float): weight on the control bound contraint
        control_reg_w (float): weight of the control regularization
        state_reg_w (float): weight of the state regularization
        terminal_velocity_w (float): weight on the terminal velocity value
        q0_reduced (vector): reduced configuration vector (base is fixed)
        v0_reduced (vector): reduced velocity vector (base is fixed)
        x0_reduced (vector): reduced initial state
        xref (vector): reference state
        uref (vector): reference control
    """

    def __init__(self, param):
        frozen_names = ["root_joint"]

        super().__init__(param, frozen_names)

        self.useFixedBase = 1

        # Cost function weights
        self.mu = 0.7
        self.foot_tracking_w = 1e4
        self.friction_cone_w = 1e3
        self.control_bound_w = 1e3
        self.control_reg_w = 1e0
        self.state_reg_w = np.array(
            [1e2] * 3 + [1e-2] * 3 + [1e2] * 6 + [1e1] * 3 + [1e0] * 3 + [1e1] * 6
        )
        self.terminal_velocity_w = np.array([0] * 12 + [1e3] * 12)

        self.q0_reduced = self.q0[7:]
        self.v0_reduced = np.zeros(self.nv)
        self.x0_reduced = np.concatenate([self.q0_reduced, self.v0_reduced])

        self.xref = self.x0_reduced
        self.uref = self.u0
