import ctypes
from ctypes import Structure
from enum import Enum
from multiprocessing import Process, Value, Array
from time import time, sleep

import numpy as np

from .WB_MPC.CasadiOCP import OCP

import quadruped_reactive_walking as qrw


class Result:
    def __init__(self, pd):
        self.xs = list(np.zeros((pd.T + 1, pd.nx)))
        self.us = list(np.zeros((pd.T, pd.nu)))
        self.acs = list(np.zeros((pd.T, pd.nv)))
        self.fs = list(np.zeros((pd.T, 12)))
        self.K = list(np.zeros([pd.T, pd.nu, pd.ndx]))
        self.solving_duration = 0.0
        self.new_result = False


class MPC_Wrapper:
    """
    Wrapper to run both types of MPC (OQSP or Crocoddyl) with the possibility to run OSQP in
    a parallel process
    """

    def __init__(self, pd, params, footsteps, gait):
        self.params = params
        self.pd = pd

        self.footsteps_plan = footsteps
        self.initial_gait = gait

        self.multiprocessing = params.enable_multiprocessing

        if self.multiprocessing:
            self.new_data = Value("b", False)
            self.running = Value("b", True)
            self.in_k = Value("i", 0)
            self.in_x0 = Array("d", [0] * pd.nx)
            self.in_warm_start = Value("b", False)
            self.in_xs = Array("d", [0] * ((pd.T + 1) * pd.nx))
            self.in_us = Array("d", [0] * (pd.T * pd.nu))
            self.in_footstep = Array("d", [0] * 12)
            self.in_gait = Array("d", [0] * ((pd.T + 1) * 4))
            self.out_xs = Array("d", [0] * ((pd.T + 1) * pd.nx))
            self.out_us = Array("d", [0] * (pd.T * pd.nu))
            self.out_k = Array("d", [0] * (pd.T * pd.nu * pd.ndx))
            self.out_solving_time = Value("d", 0.0)
        else:
            self.ocp = OCP(pd, gait)

        self.last_available_result = Result(pd)
        self.new_result = Value("b", False)

    def solve(self, k, x0, tasks, gait, guess={}):
        """
        Call either the asynchronous MPC or the synchronous MPC depending on the value
        of multiprocessing during the creation of the wrapper

        Args:
            k (int): Number of inv dynamics iterations since the start of the simulation
        """
        if self.multiprocessing:
            print("-- RUNNING ASYNCHRONOUS --")
            self.run_MPC_asynchronous(k, x0, tasks, gait, xs, us)
        else:
            print("-- RUNNING SYNCHRONOUS --")
            self.run_MPC_synchronous(x0, tasks, gait, guess)

    def get_latest_result(self):
        """
        Return the desired contact forces that have been computed by the last iteration
        of the MPC.
        If a new result is available, return the new result.
        Otherwise return the old result again.
        """
        if self.new_result.value:
            if self.multiprocessing:
                (
                    self.last_available_result.xs,
                    self.last_available_result.us,
                    self.last_available_result.K,
                    self.last_available_result.solving_duration,
                ) = self.decompress_dataOut()

            self.last_available_result.new_result = True
            self.new_result.value = False
        else:
            self.last_available_result.new_result = False

        return self.last_available_result

    def run_MPC_synchronous(self, x0, tasks, gait, guess):
        """
        Run the MPC (synchronous version)
        """

        self.ocp.solve(x0, tasks, gait, guess)
        (
            self.last_available_result.xs,
            self.last_available_result.acs,
            self.last_available_result.us,
            self.last_available_result.fs,
            self.last_available_result.solving_duration,
        ) = self.ocp.get_results()
        self.new_result.value = True

        import pinocchio as pin
        from matplotlib import pyplot as plt

        index6 = [1, 3, 5, 2, 4, 6]
        index12 = [1, 5, 9, 2, 6, 10, 3, 7, 11, 4, 8, 12]

        us = np.array(self.last_available_result.us)
        xs_raw = np.array(self.last_available_result.xs)
        xs = np.zeros((xs_raw.shape[0], xs_raw.shape[1] - 1))

        xs[:, :3] = xs_raw[:, :3]
        for i in range(xs.shape[0]):
            xs[i, 3:6] = pin.rpy.matrixToRpy(
                pin.Quaternion(xs_raw[i, 3:7]).toRotationMatrix()
            )
        xs[:, 6:] = xs_raw[:, 7:]

        plt.figure()
        N = min(10, len(self.ocp.log_vel_base))
        for k in range(N):
            plt.plot(self.ocp.log_vel_base[k])
            # plt.plot(self.ocp.log_friction_cone[k])
        plt.legend([str(k) for k in range(N)])

        plt.figure()
        plt.suptitle("Linear components")
        lgd = ["$x$", "$y$", "$z$", "$\dot{x}$", "$\dot{y}$", "$\dot{z}$"]
        for i in range(6):
            plt.subplot(3, 2, index6[i])
            if i < 3:
                plt.plot(xs[:, i])
            else:
                plt.plot(xs[:, 18 + i - 3])
            plt.legend([lgd[i]])

        plt.figure()
        plt.suptitle("Angular components")
        lgd = [
            "$\phi$",
            "$\theta$",
            "$\psi$",
            "$\dot{\omega_{x}}$",
            "$\dot{\omega_{y}}$",
            "$\dot{\omega_{z}}$",
        ]
        for i in range(6):
            plt.subplot(3, 2, index6[i])
            if i < 3:
                plt.plot(xs[:, 3 + i])
            else:
                plt.plot(xs[:, 18 + i])
            plt.legend([lgd[i]])

        plt.figure()
        plt.suptitle("Torques")
        for i in range(12):
            plt.subplot(3, 4, index12[i])
            plt.plot(us[:, i])
        plt.show()

        from IPython import embed

        embed()

    def run_MPC_asynchronous(self, k, x0, tasks, gait, xs, us):
        """
        Run the MPC (asynchronous version)
        """
        if k == 0:
            self.last_available_result.xs = [x0 for _ in range(self.pd.T + 1)]
            p = Process(target=self.MPC_asynchronous)
            p.start()

        self.add_new_data(k, x0, tasks, gait, xs, us)

    def MPC_asynchronous(self):
        """
        Parallel process with an infinite loop that run the asynchronous MPC
        """
        while self.running.value:
            if not self.new_data.value:
                continue

            self.new_data.value = False

            k, x0, tasks, gait, xs, us = self.decompress_dataIn()

            if k == 0:
                loop_ocp = OCP(self.pd, self.footsteps_plan, self.initial_gait)

            loop_ocp.solve(x0, tasks, gait, xs, us)
            xs, us, K, solving_time = loop_ocp.get_results()
            self.compress_dataOut(xs, us, K, solving_time)
            self.new_result.value = True

    def add_new_data(self, k, x0, tasks, gait, xs, us):
        """
        Compress data in a C-type structure that belongs to the shared memory to send
        data from the main control loop to the asynchronous MPC and notify the process
        that there is a new data
        """

        self.compress_dataIn(k, x0, tasks, gait, xs, us)
        self.new_data.value = True

    def compress_dataIn(self, k, x0, tasks, gait, xs, us):
        """
        Decompress data from a C-type structure that belongs to the shared memory to
        retrieve data from the main control loop in the asynchronous MPC
            dataIn (Array): shared C-type structure that contains the input data
        """
        with self.in_k.get_lock():
            self.in_k.value = k
        with self.in_x0.get_lock():
            np.frombuffer(self.in_x0.get_obj()).reshape(self.pd.nx)[:] = x0
        with self.in_footstep.get_lock():
            np.frombuffer(self.in_footstep.get_obj()).reshape((3, 4))[:, :] = tasks
        with self.in_gait.get_lock():
            np.frombuffer(self.in_gait.get_obj()).reshape((self.pd.T + 1, 4))[
                :, :
            ] = gait

        if xs is None or us is None:
            self.in_warm_start.value = False
            return

        with self.in_xs.get_lock():
            np.frombuffer(self.in_xs.get_obj()).reshape((self.pd.T + 1, self.pd.nx))[
                :, :
            ] = np.array(xs)
        with self.in_us.get_lock():
            np.frombuffer(self.in_us.get_obj()).reshape((self.pd.T, self.pd.nu))[
                :, :
            ] = np.array(us)

    def decompress_dataIn(self):
        """
        Decompress data from a C-type structure that belongs to the shared memory to
        retrieve data from the main control loop in the asynchronous MPC
        """
        with self.in_k.get_lock():
            k = self.in_k.value
        with self.in_x0.get_lock():
            x0 = np.frombuffer(self.in_x0.get_obj()).reshape(self.pd.nx)
        with self.in_footstep.get_lock():
            tasks = np.frombuffer(self.in_footstep.get_obj()).reshape((3, 4))
        with self.in_gait.get_lock():
            gait = np.frombuffer(self.in_gait.get_obj()).reshape((self.pd.T + 1, 4))

        if not self.in_warm_start.value:
            return k, x0, tasks, gait, None, None

        with self.in_xs.get_lock():
            xs = list(
                np.frombuffer(self.in_xs.get_obj()).reshape((self.pd.T + 1, self.pd.nx))
            )
        with self.in_us.get_lock():
            us = list(
                np.frombuffer(self.in_us.get_obj()).reshape((self.pd.T, self.pd.nu))
            )

        return k, x0, tasks, gait, xs, us

    def compress_dataOut(self, xs, us, K, solving_time):
        """
        Compress data to a C-type structure that belongs to the shared memory to
        retrieve data in the main control loop from the asynchronous MPC
        """
        with self.out_xs.get_lock():
            np.frombuffer(self.out_xs.get_obj()).reshape((self.pd.T + 1, self.pd.nx))[
                :, :
            ] = np.array(xs)
        with self.out_us.get_lock():
            np.frombuffer(self.out_us.get_obj()).reshape((self.pd.T, self.pd.nu))[
                :, :
            ] = np.array(us)
        with self.out_k.get_lock():
            np.frombuffer(self.out_k.get_obj()).reshape(
                [self.pd.T, self.pd.nu, self.pd.ndx]
            )[:, :, :] = np.array(K)
        self.out_solving_time.value = solving_time

    def decompress_dataOut(self):
        """
        Return the result of the asynchronous MPC (desired contact forces) that is
        stored in the shared memory
        """
        xs = list(
            np.frombuffer(self.out_xs.get_obj()).reshape((self.pd.T + 1, self.pd.nx))
        )
        us = list(np.frombuffer(self.out_us.get_obj()).reshape((self.pd.T, self.pd.nu)))
        K = list(
            np.frombuffer(self.out_k.get_obj()).reshape(
                [self.pd.T, self.pd.nu, self.pd.ndx]
            )
        )
        solving_time = self.out_solving_time.value

        return xs, us, K, solving_time

    def stop_parallel_loop(self):
        """
        Stop the infinite loop in the parallel process to properly close the simulation
        """

        self.running.value = False
